import datetime
import os

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.forms import fields
from django.core.urlresolvers import reverse

from django.db import models
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext
from django.utils.translation import ugettext_lazy as _

from django.db.models.signals import pre_save, post_delete
from django.dispatch.dispatcher import receiver

from ckeditor.fields import RichTextField
from taggit.managers import TaggableManager

from .utils import maybe_resize


LICENSES = (
    ('', _('Unspecified')),
    ('cc by', _('Creative Commons Attribution')),
    ('cc by-sa', _('Creative Commons Attribution ShareAlike')),
    ('cc by-nc', _('Creative Commons Attribution NonCommercial')),
    ('cc by-nd', _('Creative Commons Attribution NoDerivs')),
    ('cc by-nc-sa', _('Creative Commons Attribution NonCommercial ShareAlike')),
    ('cc by-nc-nd', _('Creative Commons Attribution NonCommercial NoDerivs')),
    ('cc0 / pd', _('Creative Commons Zero / Public Domain')),
    ('artlibre', _('Art Libre')),
)


class WeekdayMixin(object):
    DAY_HOUR_START = 5

    def get_weekday(self):
        weekday = self.datetime.weekday() + 7
        if self.datetime.time() < datetime.time(self.DAY_HOUR_START, 0):
            weekday -= 1
        weekday %= 7
        return weekday

    def is_on_weekday(self, day): # day is [1..7]
        week_day = self.datetime.weekday()
        if self.datetime.hour < self.DAY_HOUR_START:
            week_day -= 1
        week_day = (week_day % 7) + 1
        return week_day == day


@python_2_unicode_compatible
class Category(models.Model):

    class Meta:
        verbose_name = _('Category')
        verbose_name_plural = _('Categories')
        ordering = ['title']

    title = models.CharField(_('Title'), max_length=50)
    slug = models.SlugField(null=True)
    itunes_category = models.CharField(_('iTunes Category Name'), max_length=100, null=True)

    def sorted_emission(self):
        return self.emission_set.order_by('title')

    def __str__(self):
        return self.title


@python_2_unicode_compatible
class Colour(models.Model):

    class Meta:
        verbose_name = _('Colour')
        verbose_name_plural = _('Colours')
        ordering = ['title']

    title = models.CharField(_('Title'), max_length=50)
    slug = models.SlugField(null=True)

    def sorted_emission(self):
        return self.emission_set.order_by('title')

    def __str__(self):
        return self.title


@python_2_unicode_compatible
class Format(models.Model):

    class Meta:
        verbose_name = _('Format')
        verbose_name_plural = _('Formats')
        ordering = ['title']

    title = models.CharField(_('Title'), max_length=50)
    slug = models.SlugField(null=True)

    def __str__(self):
        return self.title


def get_emission_image_path(instance, filename):
    return os.path.join('images', instance.slug,
            os.path.basename(filename))


@python_2_unicode_compatible
class Emission(models.Model):

    class Meta:
        verbose_name = _('Emission')
        verbose_name_plural = _('Emissions')
        ordering = ['title']

    title = models.CharField(_('Title'), max_length=200)
    slug = models.SlugField(max_length=200)
    subtitle = models.CharField(_('Subtitle'), max_length=80, null=True, blank=True)
    text = RichTextField(_('Description'), null=True)
    archived = models.BooleanField(_('Archived'), default=False)
    categories = models.ManyToManyField(Category, verbose_name=_('Categories'), blank=True)
    colours = models.ManyToManyField(Colour, verbose_name=_('Colours'), blank=True)

    # XXX: languages (models.ManyToManyField(Language))

    duration = models.IntegerField(_('Duration'), default=60,
            help_text=_('In minutes'))

    default_license = models.CharField(_('Default license for podcasts'),
            max_length=20, blank=True, default='', choices=LICENSES)
    email = models.EmailField(_('Email'), max_length=254, null=True, blank=True)
    website = models.URLField(_('Website'), null=True, blank=True)

    image = models.ImageField(_('Image'),
            upload_to=get_emission_image_path, max_length=250, null=True, blank=True)

    chat_open = models.DateTimeField(null=True, blank=True)

    # denormalized from Focus
    got_focus = models.DateTimeField(default=None, null=True, blank=True)
    has_focus = models.BooleanField(default=False)

    creation_timestamp = models.DateTimeField(auto_now_add=True, null=True)
    last_update_timestamp = models.DateTimeField(auto_now=True, null=True)

    def get_absolute_url(self):
        return reverse('emission-view', kwargs={'slug': str(self.slug)})

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        super(Emission, self).save(*args, **kwargs)
        if self.id is not None and self.image:
            maybe_resize(self.image.path)

    def get_absolute_url(self):
        return reverse('emission-view',
                        kwargs={'slug':  str(self.slug)})

    def get_schedules(self):
        return Schedule.objects.filter(emission=self).order_by('datetime')

    def get_sorted_episodes(self):
        return self.episode_set.select_related().extra(select={
                'first_diffusion': 'emissions_diffusion.datetime',
                },
                select_params=(False, True),
                where=['''datetime = (SELECT MIN(datetime)
                                        FROM emissions_diffusion
                                       WHERE episode_id = emissions_episode.id)'''],
                tables=['emissions_diffusion'],
            ).order_by('-first_diffusion')

    def get_sorted_newsitems(self):
        return self.newsitem_set.select_related().order_by('-date')

    def get_next_planned_date(self, since=None):
        schedules = self.schedule_set.filter(rerun=False)
        if not schedules:
            return None
        if since is None:
            since = datetime.datetime.today()
        possible_dates = []
        for schedule in schedules:
            possible_dates.append(schedule.get_next_planned_date(since))
        possible_dates.sort()
        return possible_dates[0]


@python_2_unicode_compatible
class Schedule(models.Model, WeekdayMixin):

    class Meta:
        verbose_name = _('Schedule')
        verbose_name_plural = _('Schedules')
        ordering = ['datetime']

    WEEK_CHOICES = (
        (0b1111, _('Every week')),
        (0b0001, _('First week')),
        (0b0010, _('Second week')),
        (0b0100, _('Third week')),
        (0b1000, _('Fourth week')),
        (0b0101, _('First and third week')),
        (0b1010, _('Second and fourth week'))
    )
    emission = models.ForeignKey('Emission', verbose_name=u'Emission')
    datetime = models.DateTimeField()
    weeks = models.IntegerField(_('Weeks'), default=15, choices=WEEK_CHOICES)
    rerun = models.BooleanField(_('Rerun'), default=False)
    duration = models.IntegerField(_('Duration'), null=True, blank=True,
            help_text=_('In minutes'))

    @property
    def weeks_string(self):
        if self.weeks == 0b0001:
            return ugettext('1st of the month')
        elif self.weeks == 0b0010:
            return ugettext('2nd of the month')
        elif self.weeks == 0b0100:
            return ugettext('3rd of the month')
        elif self.weeks == 0b1000:
            return ugettext('4th of the month')
        elif self.weeks == 0b0101:
            return ugettext('1st and 3rd')
        elif self.weeks == 0b1010:
            return ugettext('2nd and 4th')
        return None

    def week_sort_key(self):
        order = [
            0b1111, # Every week
            0b0001, # First week
            0b0101, # First and third week
            0b0010, # Second week
            0b1010, # Second and fourth week
            0b0100, # Third week
            0b1000, # Fourth week
        ]
        if self.weeks in order:
            return order.index(self.weeks)
        return -1

    def get_duration(self):
        if self.duration:
            return self.duration
        else:
            return self.emission.duration

    def end_datetime(self):
        return self.datetime + datetime.timedelta(minutes=self.get_duration())

    def match_week(self, week_no):
        if week_no == 4:
            # this is the fifth week of the month, only return True for
            # emissions scheduled every week.
            return (self.weeks == 0b1111)
        if (self.weeks & (0b0001<<(week_no)) == 0):
            return False
        return True

    def matches(self, dt):
        weekday = dt.weekday()
        if dt.hour < self.DAY_HOUR_START:
            weekday -= 1
        if weekday != self.get_weekday():
            return False
        if self.weeks != 0b1111:
            week_no = (dt.day-1) // 7
            if self.match_week(week_no) is False:
                return False
        if dt.time() >= self.datetime.time() and \
                dt.time() <= (self.datetime + datetime.timedelta(minutes=self.get_duration())).time():
            return True
        return False

    def get_next_planned_date(self, since):
        possible_dates = []
        monday_since_date = since - datetime.timedelta(days=since.weekday())
        start_week_date = self.datetime.replace(
                year=monday_since_date.year,
                month=monday_since_date.month,
                day=monday_since_date.day)
        start_week_date -= datetime.timedelta(days=start_week_date.weekday())
        start_week_date += datetime.timedelta(days=self.datetime.weekday())
        for i in range(6):
            week_date = start_week_date + datetime.timedelta(days=i*7)
            if week_date < since:
                continue
            if self.match_week((week_date.day-1)//7):
                possible_dates.append(week_date)
        possible_dates.sort()
        return possible_dates[0]


    def __str__(self):
        return u'%s at %s' % (self.emission.title,
                self.datetime.strftime('%a %H:%M'))


def get_episode_image_path(instance, filename):
    return os.path.join('images', instance.emission.slug,
            os.path.basename(filename))


@python_2_unicode_compatible
class Episode(models.Model):

    class Meta:
        verbose_name = _('Episode')
        verbose_name_plural = _('Episodes')
        ordering = ['title']

    emission = models.ForeignKey('Emission', verbose_name=_('Emission'))
    title = models.CharField(_('Title'), max_length=200)
    slug = models.SlugField(max_length=200)
    subtitle = models.CharField(_('Subtitle'), max_length=150, null=True, blank=True)
    text = RichTextField(_('Description'), null=True)
    tags = TaggableManager(_('Tags'), blank=True)
    duration = models.IntegerField(_('Duration'), null=True, blank=True,
            help_text=_('In minutes'))

    image = models.ImageField(_('Image'),
            upload_to=get_episode_image_path, max_length=250, null=True, blank=True)

    effective_start = models.DateTimeField(null=True, blank=True)
    effective_end = models.DateTimeField(null=True, blank=True)

    # denormalized from Focus
    got_focus = models.DateTimeField(default=None, null=True, blank=True)
    has_focus = models.BooleanField(default=False)

    creation_timestamp = models.DateTimeField(auto_now_add=True, null=True)
    last_update_timestamp = models.DateTimeField(auto_now=True, null=True)

    # XXX: languages (models.ManyToManyField(Language))

    def __str__(self):
        return self.title

    def save(self, *args, **kwargs):
        super(Episode, self).save(*args, **kwargs)
        if self.id is not None and self.image:
            maybe_resize(self.image.path)

    def get_duration(self):
        if self.duration:
            return self.duration
        else:
            return self.emission.duration

    def get_absolute_url(self):
        return reverse('episode-view',
                        kwargs={'emission_slug': str(self.emission.slug),
                                'slug':  str(self.slug)})

    def has_sound(self):
        return (self.soundfile_set.count() > 0)

    def get_pige_download_url(self):
        return '%s/%s-%s-%s.wav' % (
                settings.PIGE_DOWNLOAD_BASE_URL,
                self.effective_start.strftime('%Y%m%d'),
                self.effective_start.strftime('%Hh%Mm%S.%f'),
                self.effective_end.strftime('%Hh%Mm%S.%f'))

    _soundfiles = {}

    @classmethod
    def set_prefetched_soundfiles(cls, soundfiles):
        cls._soundfiles.update(soundfiles)

    def has_prefetched_soundfile(self):
        return self.id in self._soundfiles

    def get_prefetched_soundfile(self):
        return self._soundfiles.get(self.id)

    _main_sound = False

    @property
    def main_sound(self):
        if self._main_sound is not False:
            return self._main_sound

        if self.has_prefetched_soundfile():
            self._main_sound = self.get_prefetched_soundfile()
            return self._main_sound

        t = self.soundfile_set.exclude(podcastable=False).exclude(fragment=True)
        if t:
            self._main_sound = t[0]
        else:
            self._main_sound = None

        return self._main_sound

    @main_sound.setter
    def main_sound(self, value):
        self._main_sound = value

    def fragment_sounds(self):
        return self.soundfile_set.exclude(podcastable=False).exclude(fragment=False)

    def diffusions(self):
        return Diffusion.objects.filter(episode=self.id).order_by('datetime')


@python_2_unicode_compatible
class Diffusion(models.Model, WeekdayMixin):

    class Meta:
        verbose_name = _('Diffusion')
        verbose_name_plural = _('Diffusions')
        ordering = ['datetime']

    episode = models.ForeignKey('Episode', verbose_name=_('Episode'))
    datetime = models.DateTimeField(_('Date/time'), db_index=True)

    def __str__(self):
        return u'%s at %02d:%02d' % (self.episode.title, self.datetime.hour,
                self.datetime.minute)

    def get_duration(self):
        return self.episode.get_duration()

    def end_datetime(self):
        return self.datetime + datetime.timedelta(minutes=self.get_duration())


@python_2_unicode_compatible
class Absence(models.Model):

    class Meta:
        verbose_name = _('Absence')
        verbose_name_plural = _('Absences')

    emission = models.ForeignKey('Emission', verbose_name=u'Emission')
    datetime = models.DateTimeField(_('Date/time'), db_index=True)

    def __str__(self):
        return u'Absence for %s on %s' % (self.emission.title, self.datetime)


def get_sound_path(instance, filename):
    return os.path.join('sounds.orig', instance.episode.emission.slug,
            os.path.basename(filename))


@python_2_unicode_compatible
class SoundFile(models.Model):

    class Meta:
        verbose_name = _('Sound file')
        verbose_name_plural = _('Sound files')
        ordering = ['creation_timestamp']

    episode = models.ForeignKey('Episode', verbose_name=_('Episode'))
    file = models.FileField(_('File'), upload_to=get_sound_path, max_length=250)
    podcastable = models.BooleanField(_('Podcastable'), default=False,
            db_index=True,
            help_text=_('The file can be published online according to SABAM rules.'))
    fragment = models.BooleanField(_('Fragment'), default=False, db_index=True,
            help_text=_('The file is some segment or extra content, not the complete recording.'))
    title = models.CharField(_('Title'), max_length=200)
    duration = models.IntegerField(_('Duration'), null=True, help_text=_('In seconds'))

    format = models.ForeignKey('Format', verbose_name=_('Format'), null=True, blank=True)
    license = models.CharField(_('License'), max_length=20, blank=True, default='', choices=LICENSES)

    # denormalized from Focus
    got_focus = models.DateTimeField(default=None, null=True, blank=True)
    has_focus = models.BooleanField(default=False)

    creation_timestamp = models.DateTimeField(auto_now_add=True, null=True)
    last_update_timestamp = models.DateTimeField(auto_now=True, null=True)

    def get_format_filename(self, format):
        return '%s_%05d__%s.%s' % (
                self.episode.slug,
                self.id,
                (self.fragment and '0' or '1'),
                format)

    def get_format_path(self, format):
        if not self.file:
            return None
        return '%s/%s' % (os.path.dirname(self.file.path).replace('.orig', ''), self.get_format_filename(format))

    def get_format_url(self, format):
        if not self.file:
            return None
        return '%s/%s' % (os.path.dirname(self.file.url).replace('.orig', ''), self.get_format_filename(format))

    def get_duration_string(self):
        if not self.duration:
            return ''
        return '%d:%02d' % (self.duration/60, self.duration%60)

    def __str__(self):
        return '%s - %s' % (self.title or self.id, self.episode.title)


@python_2_unicode_compatible
class NewsCategory(models.Model):

    class Meta:
        verbose_name = _('News Category')
        verbose_name_plural = _('News Categories')
        ordering = ['title']

    title = models.CharField(_('Title'), max_length=50)
    slug = models.SlugField(null=True)

    def __str__(self):
        return self.title

    def get_sorted_newsitems(self):
        return self.newsitem_set.select_related().order_by('-date')


def get_newsitem_image_path(instance, filename):
    return os.path.join('images', 'news', instance.slug,
            os.path.basename(filename))


@python_2_unicode_compatible
class NewsItem(models.Model):

    class Meta:
        verbose_name = _('News Item')
        verbose_name_plural = _('News Items')
        ordering = ['title']

    title = models.CharField(_('Title'), max_length=200)
    slug = models.SlugField(max_length=200)
    text = RichTextField(_('Description'))
    date = models.DateField(_('Publication Date'),
            help_text=_('The news won\'t appear on the website before this date.'))
    image = models.ImageField(_('Image'),
            upload_to=get_newsitem_image_path, max_length=250, null=True, blank=True)

    tags = TaggableManager(_('Tags'), blank=True)
    category = models.ForeignKey('NewsCategory', verbose_name=_('Category'), null=True, blank=True)
    emission = models.ForeignKey('Emission', verbose_name=_('Emission'), null=True, blank=True)

    expiration_date = models.DateField(_('Expiration Date'), null=True, blank=True)
    event_date = models.DateField(_('Event Date'), null=True, blank=True,
            help_text=_('If this is an event, set the date here so it appears in the agenda.'))

    # denormalized from Focus
    got_focus = models.DateTimeField(default=None, null=True, blank=True)
    has_focus = models.BooleanField(default=False)

    creation_timestamp = models.DateTimeField(auto_now_add=True, null=True)
    last_update_timestamp = models.DateTimeField(auto_now=True, null=True)

    def __str__(self):
        return self.title

    def get_absolute_url(self):
        return reverse('newsitem-view', kwargs={'slug': str(self.slug)})


@python_2_unicode_compatible
class Nonstop(models.Model):

    class Meta:
        verbose_name = _('Nonstop zone')
        verbose_name_plural = _('Nonstop zones')
        ordering = ['title']

    title = models.CharField(_('Title'), max_length=50)
    slug = models.SlugField()

    start = models.TimeField(_('Start'))
    end = models.TimeField(_('End'))
    text = RichTextField(_('Description'), null=True, blank=True)
    redirect_path = models.CharField(_('Redirect Path'), max_length=200, blank=True)

    def __str__(self):
        return self.title


@python_2_unicode_compatible
class Focus(models.Model):
    title = models.CharField(_('Alternate Title'), max_length=50,
            null=True, blank=True)
    newsitem = models.ForeignKey('NewsItem', verbose_name=_('News Item'),
            null=True, blank=True)
    emission = models.ForeignKey('Emission', verbose_name=_('Emission'),
            null=True, blank=True)
    episode = models.ForeignKey('Episode', verbose_name=_('Episode'),
            null=True, blank=True)
    soundfile = models.ForeignKey('SoundFile', verbose_name=_('Sound file'),
            null=True, blank=True)
    page = models.ForeignKey('data.Page', verbose_name=_('Page'),
            null=True, blank=True)
    current = models.BooleanField('Current', default=True)

    creation_timestamp = models.DateTimeField(auto_now_add=True, null=True)
    last_update_timestamp = models.DateTimeField(auto_now=True, null=True)

    def __str__(self):
        if self.newsitem:
            return u'Newsitem: %s' % self.newsitem.title
        if self.emission:
            return u'Emission: %s' % self.emission.title
        if self.episode:
            return u'Episode: %s' % self.episode.title
        if self.soundfile:
            return u'Soundfile: %s' % (self.soundfile.title or self.soundfile.episode.title)
        if self.page:
            return u'Page: %s' % self.page.title
        return u'%s' % self.id

    def focus_title(self):
        if self.title:
            return self.title
        if self.newsitem:
            return self.newsitem.title
        if self.emission:
            return self.emission.title
        if self.page:
            return self.page.title

        episode = None
        if self.soundfile:
            if self.soundfile.fragment and self.soundfile.title:
                return self.soundfile.title
            episode = self.soundfile.episode
        elif self.episode:
            episode = self.episode

        if episode:
            if episode.title:
                return episode.title
            else:
                return episode.emission.title

        return None

    def content_image(self):
        if self.newsitem:
            return self.newsitem.image
        if self.emission:
            return self.emission.image
        if self.page:
            from panikombo.models import Topik
            return Topik.objects.get(page=self.page).image

        episode = None
        if self.soundfile:
            episode = self.soundfile.episode
        elif self.episode:
            episode = self.episode

        if episode:
            if episode.image:
                return episode.image
            else:
                return episode.emission.image

    def content_category_title(self):
        if self.newsitem:
            if self.newsitem.category:
                return self.newsitem.category.title
            return _('News')
        if self.emission:
            return _('Emission')
        if self.episode:
            return self.episode.emission.title
        if self.soundfile:
            return self.soundfile.episode.emission.title
        if self.page:
            return 'Topik'

    def get_related_object(self):
        try:
            if self.newsitem:
                return self.newsitem
            if self.emission:
                return self.emission
            if self.episode:
                return self.episode
            if self.soundfile:
                return self.soundfile
            if self.page:
                from panikombo.models import Topik
                return Topik.objects.get(page=self.page)
        except ObjectDoesNotExist:
            return None

        return None


def get_playlist_sound_path(instance, filename):
    return os.path.join('playlists', instance.episode.emission.slug,
            instance.episode.slug, os.path.basename(filename))


class PlaylistElement(models.Model):
    episode = models.ForeignKey('Episode', null=True)
    title = models.CharField(_('Title'), max_length=200)
    notes = models.CharField(_('Notes'), max_length=200, blank=True, default='')
    sound = models.FileField(_('Sound'), upload_to=get_playlist_sound_path, max_length=250)
    order = models.PositiveIntegerField()

    class Meta:
        verbose_name = _('Playlist Element')
        verbose_name_plural = _('Playlist Elements')
        ordering = ['order']

    def shortcut(self):
        return chr(ord('a')+self.order-1)


@receiver(pre_save, sender=Focus, dispatch_uid='focus_pre_save')
def set_focus_on_save(sender, instance, **kwargs):
    object = instance.get_related_object()
    must_save = False
    if instance.current != object.has_focus:
        object.has_focus = instance.current
        must_save = True
    if object and not object.got_focus and instance.current:
        object.got_focus = datetime.datetime.now()
        must_save = True
    if must_save:
        object.save()

@receiver(post_delete, sender=Focus, dispatch_uid='focus_post_delete')
def remove_focus_on_delete(sender, instance, **kwargs):
    object = instance.get_related_object()
    if object and (object.got_focus or object.has_focus):
        object.got_focus = None
        object.has_focus = False
        object.save()
