import datetime
import json
import os
import time

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.core.files.storage import DefaultStorage
from django.core.urlresolvers import reverse_lazy
from django.http import HttpResponse, Http404, HttpResponseRedirect
from django.core.urlresolvers import reverse, reverse_lazy
from django.shortcuts import redirect
from django.utils.translation import ugettext as _, ugettext_lazy

from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import View, TemplateView, RedirectView
from django.views.generic.edit import CreateView, UpdateView, DeleteView
from django.views.generic.list import ListView
from django.views.generic.detail import DetailView

from django.contrib import messages

import haystack.views

from .models import Emission, Episode, Diffusion, Category, Schedule, \
                    SoundFile, NewsItem, NewsCategory, Absence, \
                    PlaylistElement
from .forms import EmissionForm, EpisodeForm, EpisodeNewForm, ScheduleForm, \
                   DiffusionForm, SoundFileForm, NewsItemForm, \
                   SoundFileEditForm, AbsenceForm, PlaylistElementForm


__all__ = ['EmissionListView', 'EmissionDetailView', 'EmissionCreateView',
           'EmissionUpdateView', 'EmissionDeleteView',
           'EpisodeCreateView', 'EpisodeDetailView', 'EpisodeUpdateView',
           'EpisodeDeleteView', 'EmissionAddScheduleView',
           'ScheduleDeleteView', 'CategoryListView', 'DaysView',
           'UploadView', 'EpisodeAddSoundFileView',
           'EpisodeAddDiffusionView', 'DiffusionDeleteView',
           'EmissionNewsItemAdd', 'NewsItemDetailView',
           'NewsItemUpdateView', 'CategoryNewsItemAddView',
           'NewsItemDeleteView', 'NewsItemAddView',
           'SoundFileDeleteView', 'SoundFileUpdateView',
           'EmissionAddAbsenceView', 'AbsenceDeleteView',
           'EmissionOpenChatView', 'EmissionCloseChatView',
           'EpisodeRegieView',
           'EpisodeRegieUpdateOrderView',
           'EpisodeRegieMarks', 'EpisodeRegieDeleteElementView',
           'EpisodeStartView', 'EpisodeStopView',
           ]


SUCCESS_MESSAGE = ugettext_lazy('Your changes will appear online in a few minutes.')


class EmissionListView(ListView):
    model = Emission

    def get_queryset(self):
        return Emission.objects.order_by('title')


class CategoryListView(ListView):
    model = Category

    def get_queryset(self):
        return Category.objects.order_by('title')


class EmissionDetailView(DetailView):
    model = Emission

    def get_context_data(self, **kwargs):
        context = super(EmissionDetailView, self).get_context_data(**kwargs)
        context['add_schedule_form'] = ScheduleForm(initial={'emission': self.object})
        context['add_absence_form'] = AbsenceForm(
                initial={'emission': self.object, 'user': self.request.user})
        context['schedules'] = Schedule.objects.select_related().filter(
                emission=self.object).order_by('datetime')

        # get all episodes, with an additional attribute to get the date of
        # their first diffusion
        context['episodes'] = \
                Episode.objects.select_related().filter(emission=self.object
                    ).extra(select={
                        'first_diffusion': 'emissions_diffusion.datetime',
                        },
                        select_params=(False, True),
                        where=['''datetime = (SELECT MIN(datetime)
                                                FROM emissions_diffusion
                                               WHERE episode_id = emissions_episode.id)'''],
                        tables=['emissions_diffusion'],
                    ).order_by('-first_diffusion').distinct()

        # get all related soundfiles in a single query
        soundfiles = {}
        for soundfile in SoundFile.objects.filter(podcastable=True,
                fragment=False, episode__emission=self.object):
            soundfiles[soundfile.episode_id] = soundfile

        Episode.set_prefetched_soundfiles(soundfiles)

        context['newsitems'] = self.object.get_sorted_newsitems()
        context['absences'] = self.object.absence_set.select_related(
                ).filter(datetime__gte=datetime.date.today()).order_by('datetime')

        if not 'all' in self.request.GET:
            context['episodes'] = context['episodes'][:10]
            if context['absences']:
                context['newsitems'] = context['newsitems'][:5]
            else:
                context['newsitems'] = context['newsitems'][:10]

        try:
            context['can_manage'] = self.request.user.can_manage(self.object)
        except AttributeError:
            pass

        return context


class EmissionCreateView(CreateView):
    form_class = EmissionForm
    model = Emission

    success_url = reverse_lazy('emission-list')

    def get_form(self, *args, **kwargs):
        if not self.request.user.has_perm('emissions.add_emission'):
            raise PermissionDenied()
        return super(EmissionCreateView, self).get_form(*args, **kwargs)

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return super(EmissionCreateView, self).get_success_url()


class EmissionUpdateView(UpdateView):
    form_class = EmissionForm
    model = Emission

    def get_form(self, *args, **kwargs):
        if not self.request.user.can_manage(self.object):
            raise PermissionDenied()
        return super(EmissionUpdateView, self).get_form(*args, **kwargs)

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return super(EmissionUpdateView, self).get_success_url()


class EmissionDeleteView(DeleteView):
    model = Emission
    success_url = reverse_lazy('home')

    def get_form(self, *args, **kwargs):
        if not self.request.user.can_manage(self.object):
            raise PermissionDenied()
        return super(EmissionDeleteView, self).get_form(*args, **kwargs)

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return super(EmissionDeleteView, self).get_success_url()


class EpisodeCreateView(CreateView):
    model = Episode
    form_class = EpisodeNewForm

    def get_form(self, *args, **kwargs):
        emission = Emission.objects.get(slug=self.kwargs.get('emission_slug'))
        if not self.request.user.can_manage(emission):
            raise PermissionDenied()
        return super(EpisodeCreateView, self).get_form(*args, **kwargs)

    def get_initial(self):
        initial = super(EpisodeCreateView, self).get_initial()
        initial['emission'] = Emission.objects.get(slug=self.kwargs.get('emission_slug'))
        initial['duration'] = initial['emission'].duration
        initial['first_diffusion'] = initial['emission'].get_next_planned_date()
        for i, schedule in enumerate(
                Schedule.objects.filter(emission=initial['emission'],
                    rerun=True).order_by('datetime')):
            rerun_date = schedule.get_next_planned_date(initial['first_diffusion'])
            if i == 0:
                initial['second_diffusion'] = rerun_date
            elif i == 1:
                initial['third_diffusion'] = rerun_date
            elif i == 2:
                initial['fourth_diffusion'] = rerun_date
        return initial

    def get_context_data(self, **kwargs):
        context = super(EpisodeCreateView, self).get_context_data(**kwargs)
        context['emission'] = Emission.objects.get(slug=self.kwargs.get('emission_slug'))
        return context

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return self.object.get_absolute_url()


class EmissionEpisodeMixin(object):
    def get_queryset(self):
        return self.model.objects.filter(emission__slug=self.kwargs['emission_slug'])


class EpisodeDetailView(EmissionEpisodeMixin, DetailView):
    model = Episode

    def get_context_data(self, **kwargs):
        context = super(EpisodeDetailView, self).get_context_data(**kwargs)
        context['diffusions'] = Diffusion.objects.filter(episode=self.object.id)
        context['soundfiles'] = SoundFile.objects.filter(episode=self.object.id)

        try:
            context['can_manage'] = self.request.user.can_manage(self.object.emission)
        except AttributeError:
            pass

        return context


class EpisodeUpdateView(EmissionEpisodeMixin, UpdateView):
    form_class = EpisodeForm
    model = Episode

    def get_form(self, *args, **kwargs):
        if not self.request.user.can_manage(self.object.emission):
            raise PermissionDenied()
        return super(EpisodeUpdateView, self).get_form(*args, **kwargs)

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return super(EpisodeUpdateView, self).get_success_url()


class EpisodeDeleteView(EmissionEpisodeMixin, DeleteView):
    model = Episode

    def get_form(self, *args, **kwargs):
        if not self.request.user.can_manage(self.object.emission):
            raise PermissionDenied()
        return super(EpisodeDeleteView, self).get_form(*args, **kwargs)

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return reverse('emission-view', kwargs={
            'slug': self.object.emission.slug})


class EmissionAddScheduleView(CreateView):
    form_class = ScheduleForm
    model = Schedule

    def get_form(self, *args, **kwargs):
        if not self.request.user.has_perm('emissions.add_schedule'):
            raise PermissionDenied()
        return super(EmissionAddScheduleView, self).get_form(*args, **kwargs)

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return self.object.emission.get_absolute_url()


class ScheduleDeleteView(RedirectView):
    def get_redirect_url(self, emission_slug, pk):
        if not self.request.user.has_perm('emissions.delete_schedule'):
            raise PermissionDenied()
        Schedule.objects.filter(id=pk).delete()
        messages.success(self.request, SUCCESS_MESSAGE)
        return reverse('emission-view', kwargs={'slug': str(emission_slug)})


class EmissionAddAbsenceView(CreateView):
    form_class = AbsenceForm
    model = Absence

    def get_initial(self):
        initial = super(EmissionAddAbsenceView, self).get_initial()
        initial['emission'] = Emission.objects.get(slug=self.kwargs.get('slug'))
        initial['user'] = self.request.user
        return initial

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return self.object.emission.get_absolute_url()

    def get_form(self, *args, **kwargs):
        emission = Emission.objects.get(slug=self.kwargs.get('slug'))
        if not self.request.user.can_manage(emission):
            raise PermissionDenied()
        return super(EmissionAddAbsenceView, self).get_form(*args, **kwargs)


class AbsenceDeleteView(RedirectView):
    def get_redirect_url(self, emission_slug, pk):
        if not self.request.user.has_perm('emissions.delete_absence'):
            raise PermissionDenied()
        Absence.objects.filter(id=pk).delete()
        messages.success(self.request, SUCCESS_MESSAGE)
        return reverse('emission-view', kwargs={'slug': str(emission_slug)})


class DaysView(TemplateView):
    template_name = 'emissions/days.html'

    def get_context_data(self, **kwargs):
        context = super(DaysView, self).get_context_data(**kwargs)
        schedules = Schedule.objects.all().order_by('datetime')
        days = []
        for day in range(7):
            days.append({'schedules': [x for x in schedules if x.is_on_weekday(day+1)],
                         'datetime': datetime.datetime(2007, 1, day+1)})
        context['days'] = days
        return context

class JSONResponse(HttpResponse):
    """JSON response class."""
    def __init__(self, obj='', json_opts={}, mimetype='application/json', *args, **kwargs):
        content = json.dumps(obj, **json_opts)
        super(JSONResponse,self).__init__(content, mimetype, *args, **kwargs)


class UploadView(View):

    def response_mimetype(self, request):
        if 'application/json' in request.META['HTTP_ACCEPT']:
            return 'application/json'
        else:
            return 'text/plain'

    @csrf_exempt
    def post(self, request, transaction_id):
        storage = DefaultStorage()
        max_filename_length = 256
        url = reverse('upload', kwargs={'transaction_id': transaction_id})
        if request.FILES is None:
            response = JSONResponse({}, {}, self.response_mimetype(request))
            response['Content-Disposition'] = 'inline; filename=files.json'
            return response
        data = []
        for uploaded_file in request.FILES.values():
            path = os.path.join('upload', str(transaction_id), uploaded_file.name)
            filename = storage.save(path, uploaded_file)
            url = '%s%s' % (url, os.path.basename(filename))
            data.append({'name': uploaded_file.name, 'size': uploaded_file.size, 'url': url})
        response = JSONResponse(data, {}, self.response_mimetype(request))
        response['Content-Disposition'] = 'inline; filename=files.json'
        return response


class EpisodeAddSoundFileView(CreateView):
    form_class = SoundFileForm
    model = SoundFile

    def get_initial(self):
        context = {
            'episode': Episode.objects.get(
                slug=self.kwargs.get('slug'),
                emission__slug=self.kwargs.get('emission_slug')),
            'title': _('Record'),
        }
        context['license'] = context['episode'].emission.default_license
        return context

    def get_form(self, *args, **kwargs):
        emission = Emission.objects.get(slug=self.kwargs.get('emission_slug'))
        if not self.request.user.can_manage(emission):
            raise PermissionDenied()
        return super(EpisodeAddSoundFileView, self).get_form(*args, **kwargs)

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return self.object.episode.get_absolute_url()


class EpisodeAddDiffusionView(CreateView):
    form_class = DiffusionForm
    model = Diffusion

    def get_initial(self):
        return {
            'episode': Episode.objects.get(
                slug=self.kwargs.get('slug'),
                emission__slug=self.kwargs.get('emission_slug')),
        }

    def get_form(self, *args, **kwargs):
        if not self.request.user.has_perm('emissions.add_diffusion'):
            raise PermissionDenied()
        episode = Episode.objects.get(slug=self.kwargs.get('slug'),
                emission__slug=self.kwargs.get('emission_slug'))
        if not self.request.user.can_manage(episode):
            raise PermissionDenied()
        return super(EpisodeAddDiffusionView, self).get_form(*args, **kwargs)

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return self.object.episode.get_absolute_url()


class DiffusionDeleteView(RedirectView):
    def get_redirect_url(self, emission_slug, slug, pk):
        if not self.request.user.has_perm('emissions.delete_diffusion'):
            raise PermissionDenied()
        episode = Episode.objects.get(slug=slug, emission__slug=emission_slug)
        if not self.request.user.can_manage(episode):
            raise PermissionDenied()
        Diffusion.objects.filter(id=pk).delete()
        messages.success(self.request, SUCCESS_MESSAGE)
        return reverse('episode-view', kwargs={'emission_slug': str(emission_slug),
                                               'slug': str(slug)})


class FacetedSearchView(haystack.views.FacetedSearchView):
    def extra_context(self):
        context = super(FacetedSearchView, self).extra_context()
        context['selected_categories'] = [
                x.split(':', 1)[1] for x in self.request.GET.getlist('selected_facets')
                if x.startswith('categories_exact')]
        context['selected_tags'] = [
                x.split(':', 1)[1] for x in self.request.GET.getlist('selected_facets')
                if x.startswith('tags_exact')]
        return context


class EmissionNewsItemAdd(CreateView):
    model = NewsItem
    form_class = NewsItemForm

    def get_form(self, *args, **kwargs):
        emission = Emission.objects.get(slug=self.kwargs.get('emission_slug'))
        if not self.request.user.can_manage(emission):
            raise PermissionDenied()
        return super(EmissionNewsItemAdd, self).get_form(*args, **kwargs)

    def get_initial(self):
        initial = super(EmissionNewsItemAdd, self).get_initial()
        initial['emission'] = Emission.objects.get(slug=self.kwargs.get('emission_slug'))
        initial['date'] = datetime.datetime.today()
        initial['user'] = self.request.user
        return initial

    def get_context_data(self, **kwargs):
        context = super(EmissionNewsItemAdd, self).get_context_data(**kwargs)
        context['emission'] = Emission.objects.get(slug=self.kwargs.get('emission_slug'))
        return context

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return self.object.get_absolute_url()


class NewsItemDetailView(DetailView):
    model = NewsItem

    def get_context_data(self, **kwargs):
        context = super(NewsItemDetailView, self).get_context_data(**kwargs)
        context['can_manage'] = self.request.user.can_manage(self.object)
        return context


class NewsItemUpdateView(UpdateView):
    form_class = NewsItemForm
    model = NewsItem

    def get_form(self, *args, **kwargs):
        if not self.request.user.can_manage(self.object):
            raise PermissionDenied()
        return super(NewsItemUpdateView, self).get_form(*args, **kwargs)

    def get_initial(self):
        initial = super(NewsItemUpdateView, self).get_initial()
        initial['user'] = self.request.user
        return initial

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return super(NewsItemUpdateView, self).get_success_url()


class CategoryNewsItemAddView(CreateView):
    model = NewsItem
    form_class = NewsItemForm

    def get_form(self, *args, **kwargs):
        if not self.request.user.has_perm('emissions.add_newsitem'):
            raise PermissionDenied()
        return super(CategoryNewsItemAddView, self).get_form(*args, **kwargs)

    def get_initial(self):
        initial = super(CategoryNewsItemAddView, self).get_initial()
        initial['category'] = NewsCategory.objects.get(slug=self.kwargs.get('slug'))
        initial['date'] = datetime.datetime.today()
        initial['user'] = self.request.user
        return initial

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return self.object.get_absolute_url()


class NewsItemAddView(CreateView):
    model = NewsItem
    form_class = NewsItemForm

    def get_form(self, *args, **kwargs):
        if not self.request.user.has_perm('emissions.add_newsitem'):
            raise PermissionDenied()
        return super(NewsItemAddView, self).get_form(*args, **kwargs)

    def get_initial(self):
        initial = super(NewsItemAddView, self).get_initial()
        initial['date'] = datetime.datetime.today()
        initial['user'] = self.request.user
        return initial

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return self.object.get_absolute_url()


class NewsItemDeleteView(DeleteView):
    model = NewsItem
    success_url = reverse_lazy('home')

    def get_form(self, *args, **kwargs):
        if not self.request.user.can_manage(self.object):
            raise PermissionDenied()
        return super(NewsItemDeleteView, self).get_form(*args, **kwargs)

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return super(NewsItemDeleteView, self).get_success_url()


class SoundFileDeleteView(DeleteView):
    model = SoundFile

    def get_form(self, *args, **kwargs):
        if not self.request.user.can_manage(self.object):
            raise PermissionDenied()
        return super(SoundFileDeleteView, self).get_form(*args, **kwargs)

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return reverse('episode-view', kwargs={
            'emission_slug': self.object.episode.emission.slug,
            'slug': self.object.episode.slug})


class SoundFileUpdateView(UpdateView):
    template_name_suffix = '_form_update'
    form_class = SoundFileEditForm
    model = SoundFile

    def get_form(self, *args, **kwargs):
        if not self.request.user.can_manage(self.object.episode.emission):
            raise PermissionDenied()
        return super(SoundFileUpdateView, self).get_form(*args, **kwargs)

    def get_success_url(self):
        messages.success(self.request, SUCCESS_MESSAGE)
        return '../../'


class EmissionOpenChatView(DetailView):
    model = Emission

    def render_to_response(self, context):
        self.object.chat_open = datetime.datetime.now()
        self.object.save()
        return redirect(self.object.get_absolute_url())


class EmissionCloseChatView(DetailView):
    model = Emission

    def render_to_response(self, context):
        self.object.chat_open = None
        self.object.save()
        return redirect(self.object.get_absolute_url())


class EpisodeRegieView(EmissionEpisodeMixin, DetailView):
    model = Episode
    template_name = 'emissions/episode_regie.html'

    def get_object(self, queryset=None):
        try:
            return Episode.objects.get(slug=self.kwargs.get('slug'),
                    emission__slug=self.kwargs.get('emission_slug'))
        except Episode.DoesNotExist:
            raise Http404()

    def post(self, request, *args, **kwargs):
        episode = self.get_object()
        form = PlaylistElementForm(request.POST, request.FILES)
        if not form.is_valid():
            messages.error(request, _('Error adding file'))
            return HttpResponseRedirect('.')
        i = PlaylistElement.objects.filter(episode=episode).count()
        playlist_element = PlaylistElement(
                episode=episode,
                title=os.path.basename(form.cleaned_data['sound'].name),
                sound=form.cleaned_data['sound'],
                order=i+1)
        playlist_element.save()
        messages.info(request, _('File uploaded successfully.'))
        return HttpResponseRedirect('.')

    def get_context_data(self, **kwargs):
        context = super(EpisodeRegieView, self).get_context_data(**kwargs)
        context['playlist'] = PlaylistElement.objects.filter(episode=self.object.id)
        context['start_time'] = ''
        context['end_time'] = ''
        if self.object.effective_start:
            context['ready'] = 'ready'
            context['start_time'] = int(time.mktime(self.object.effective_start.timetuple())*1000
                    + self.object.effective_start.microsecond/1000)
        if self.object.effective_end:
            context['end_time'] = int(time.mktime(self.object.effective_end.timetuple())*1000
                    + self.object.effective_end.microsecond/1000)
            context['download_url'] = self.object.get_pige_download_url()

        try:
            context['can_manage'] = self.request.user.can_manage(self.object)
        except AttributeError:
            pass

        context['upload_file_form'] = PlaylistElementForm(initial={'episode': self.object})

        return context


class EpisodeRegieUpdateOrderView(View):
    def get(self, request, *args, **kwargs):
        new_order = request.GET.getlist('new-order[]')
        for element in PlaylistElement.objects.filter(id__in=new_order):
            element.order = new_order.index(str(element.id)) + 1
            element.save()
        return HttpResponse('ok')


class EpisodeRegieDeleteElementView(EpisodeRegieView):
    def get(self, request, *args, **kwargs):
        element_id = request.GET.get('id')
        element = PlaylistElement.objects.get(id=element_id)
        if not self.request.user.can_manage(element.episode):
            return JSONResponse({'err': 1})
        element.delete()
        return HttpResponseRedirect('..')


class EpisodeRegieMarks(EpisodeRegieView):
    def get(self, request, *args, **kwargs):
        episode = self.get_object()
        start = self.request.GET['start']
        end = self.request.GET['end']
        if start:
            episode.effective_start = datetime.datetime.fromtimestamp(float(start)/1000)
        else:
            episode.effective_start = None
        if end:
            episode.effective_end = datetime.datetime.fromtimestamp(float(end)/1000)
        else:
            episode.effective_end = None
        episode.save()
        obj = {'err': 0}
        if end:
            obj['pige_download_url'] = episode.get_pige_download_url()
        return JSONResponse(obj)


class EpisodeStartView(EmissionEpisodeMixin, RedirectView):
    def get_redirect_url(self, emission_slug, slug):
        episode = Episode.objects.get(slug=slug, emission__slug=emission_slug)
        episode.effective_start = datetime.datetime.now()
        episode.save()
        messages.success(self.request, _('Started recording'))
        return reverse('episode-view',
                kwargs={'emission_slug': emission_slug,
                        'slug': slug})


class EpisodeStopView(EmissionEpisodeMixin, RedirectView):
    def get_redirect_url(self, emission_slug, slug):
        episode = Episode.objects.get(slug=slug, emission__slug=emission_slug)
        episode.effective_end = datetime.datetime.now()
        episode.save()
        messages.success(self.request, _('Stopped recording'))
        return reverse('episode-view',
                kwargs={'emission_slug': emission_slug,
                        'slug': slug})
